# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

"""Model a varlink interface method and its type."""

import collections.abc
import dataclasses
import functools
import inspect
import typing

from .conversion import ObjectVarlinkType, VarlinkType, _merge_typedefs
from .error import GenericVarlinkErrorReply, TypedVarlinkErrorReply
from .message import VarlinkMethodReply
from .types import JSONObject, validate_interface


_P = typing.ParamSpec("_P")
_R = typing.TypeVar("_R")


@dataclasses.dataclass
class VarlinkMethodSignature:
    """Annotate methods with such a signature object to indicate that they are
    meant to be called via varlink. The signature contains prepared information
    about the parameter conversion and how to call it.
    """

    asynchronous: bool
    """Indicates whether the method is async or not."""

    more: bool
    """Indicates whether the method returns and iterable or not."""

    parameter_type: ObjectVarlinkType
    """A type for the call parameters."""

    return_type: ObjectVarlinkType
    """A type for the return parameters."""


@dataclasses.dataclass
class AnnotatedResult:
    """A wrapper for unconverted varlink reply values to affect protocol-level
    details.
    """

    parameters: collections.abc.Mapping[str, typing.Any]
    """Parameter mapping that would normally returned without this wrapper."""

    _: dataclasses.KW_ONLY

    continues: bool = False
    """Indicates whether another reply follows this reply. Setting this is only
    valid if the caller enabled the "more" flag. If setting this to True,
    another reply must follow. Otherwise, this must be the last reply.
    """

    extensions: JSONObject = dataclasses.field(default_factory=dict)
    """May be used to supply extension fields on the varlink reply level."""


class LastResult(Exception):
    """An varlink-specific exception for signalling a non-continued reply after
    a sequence of continued replies."""

    def __init__(self, value: typing.Any):
        """Return the given value as the last reply."""
        if isinstance(value, AnnotatedResult) and value.continues:
            raise RuntimeError(
                "An AnnotatedResult raised via LastResult cannot continue."
            )
        self.value = value


_MethodResultType = (
    AnnotatedResult
    | collections.abc.Awaitable[AnnotatedResult]
    | collections.abc.Iterator[AnnotatedResult]
    | collections.abc.AsyncIterator[AnnotatedResult]
)


def _params_to_varlinkobj(
    params: collections.abc.Iterator[tuple[str, inspect.Parameter]],
) -> ObjectVarlinkType:
    return ObjectVarlinkType(
        {
            name: (
                VarlinkType.from_type_annotation(tobj.annotation)
                if tobj.default is tobj.empty
                else VarlinkType.from_type_annotation(
                    tobj.annotation
                ).optional()
            )
            for name, tobj in params
        },
    )


class _VarlinkMethodDecorator(typing.Protocol):
    @typing.overload
    def __call__(
        self,
        function: collections.abc.Callable[
            _P, collections.abc.AsyncIterator[_R]
        ],
    ) -> collections.abc.Callable[
        _P, collections.abc.AsyncIterator[AnnotatedResult]
    ]: ...

    @typing.overload
    def __call__(
        self,
        function: collections.abc.Callable[_P, collections.abc.Iterator[_R]],
    ) -> collections.abc.Callable[
        _P, collections.abc.Iterator[AnnotatedResult]
    ]: ...

    @typing.overload
    def __call__(
        self,
        function: collections.abc.Callable[_P, collections.abc.Awaitable[_R]],
    ) -> collections.abc.Callable[
        _P, collections.abc.Awaitable[AnnotatedResult]
    ]: ...

    @typing.overload
    def __call__(
        self, function: collections.abc.Callable[_P, _R]
    ) -> collections.abc.Callable[_P, _MethodResultType]: ...


@typing.overload
def varlinkmethod(
    function: collections.abc.Callable[_P, collections.abc.AsyncIterator[_R]],
    *,
    return_parameter: str | None = None,
    delay_generator: bool = True,
) -> collections.abc.Callable[
    _P, collections.abc.AsyncIterator[AnnotatedResult]
]: ...


@typing.overload
def varlinkmethod(
    function: collections.abc.Callable[_P, collections.abc.Iterator[_R]],
    *,
    return_parameter: str | None = None,
    delay_generator: bool = True,
) -> collections.abc.Callable[
    _P, collections.abc.Iterator[AnnotatedResult]
]: ...


@typing.overload
def varlinkmethod(
    function: collections.abc.Callable[_P, collections.abc.Awaitable[_R]],
    *,
    return_parameter: str | None = None,
    delay_generator: bool = True,
) -> collections.abc.Callable[
    _P, collections.abc.Awaitable[AnnotatedResult]
]: ...


@typing.overload
def varlinkmethod(
    function: collections.abc.Callable[_P, _R],
    *,
    return_parameter: str | None = None,
    delay_generator: bool = True,
) -> collections.abc.Callable[_P, _MethodResultType]: ...


@typing.overload
def varlinkmethod(
    *, return_parameter: str | None = None, delay_generator: bool = True
) -> _VarlinkMethodDecorator: ...


# Whilst the Python documentation says the implementation should be untyped,
# mypy insists on having a type to type check the body of the function.
# https://github.com/python/mypy/issues/3360
def varlinkmethod(
    function: collections.abc.Callable[_P, _R] | None = None,
    *,
    return_parameter: str | None = None,
    delay_generator: bool = True,
) -> (
    collections.abc.Callable[
        [collections.abc.Callable[_P, _R]],
        collections.abc.Callable[_P, _MethodResultType],
    ]
    | collections.abc.Callable[_P, _MethodResultType]
):
    """Decorator for fully type annotated methods to be callable from varlink.
    The function may be a generator, in which case it should be called
    with the "more" field set on the varlink side.

    The function has multiple options to return or yield values. It may produce
    AnnotatedResult instances. If return_parameter is None, it may produce a
    dict and a bare value to be wrapped in a dict with return_parameter as key
    otherwise.

    If the function is a generator (async or not), AnnotatedResult are
    forwarded immediately. They must correctly indicate whether the generator
    produces another element by setting the continues field appropriately. If
    delay_generator is True, other values will be delayed until the next
    element is produced (and thus the continues field for the previous element
    is known). Otherwise, all values are forwarded immediately assuming that
    generator never finishes via StopIteration or AsyncStopIteration. In the
    latter case, a final value may produced by raising it inside as LastResult
    exception. The function must produce at least one result or raise as
    LastResult exception.
    """

    def wrap(
        function: collections.abc.Callable[_P, _R],
    ) -> collections.abc.Callable[_P, AnnotatedResult]:
        asynchronous = inspect.iscoroutinefunction(function)
        asyncgen = inspect.isasyncgenfunction(function)
        signature = inspect.signature(function)
        param_iterator = iter(signature.parameters.items())
        if next(param_iterator)[0] != "self":
            raise RuntimeError(
                "first argument of a method should be named self"
            )
        return_type = signature.return_annotation
        more = False
        ret_origin = typing.get_origin(return_type)
        if ret_origin is not None and issubclass(
            ret_origin,
            (
                collections.abc.AsyncIterator
                if asyncgen
                else collections.abc.Iterator
            ),
        ):
            return_type = typing.get_args(return_type)[0]
            more = True
        return_vtype = VarlinkType.from_type_annotation(return_type)
        make_result: collections.abc.Callable[[_R], AnnotatedResult]
        if return_parameter is not None:
            return_vtype = ObjectVarlinkType({return_parameter: return_vtype})

            def make_result(result: _R) -> AnnotatedResult:
                return AnnotatedResult({return_parameter: result})

        elif return_type is None:
            return_vtype = ObjectVarlinkType({})

            def make_result(_result: _R) -> AnnotatedResult:
                return AnnotatedResult({})

        elif not isinstance(return_vtype, ObjectVarlinkType):
            raise TypeError("a varlinkmethod must return a mapping")
        else:
            # mypy cannot figure out that a type also is a Callable.
            make_result = typing.cast(
                collections.abc.Callable[[_R], AnnotatedResult],
                AnnotatedResult,
            )

        vlsig = VarlinkMethodSignature(
            asyncgen or asynchronous,
            more,
            _params_to_varlinkobj(param_iterator),
            return_vtype,
        )
        wrapped: collections.abc.Callable[_P, typing.Any]
        if more and asyncgen:
            asynciterfunction = typing.cast(
                collections.abc.Callable[
                    _P, collections.abc.AsyncIterable[_R]
                ],
                function,
            )

            if delay_generator:

                @functools.wraps(function)
                async def wrapped(
                    *args: _P.args, **kwargs: _P.kwargs
                ) -> collections.abc.AsyncGenerator[AnnotatedResult, None]:
                    pending = None
                    try:
                        async for result in asynciterfunction(*args, **kwargs):
                            if pending is not None:
                                pending.continues = True
                                yield pending
                                pending = None
                            if isinstance(result, AnnotatedResult):
                                yield result
                            else:
                                pending = make_result(result)
                    except LastResult as exc:
                        if pending is not None:
                            pending.continues = True
                            yield pending
                        if isinstance(exc.value, AnnotatedResult):
                            yield exc.value
                        else:
                            yield make_result(exc.value)
                    except:
                        if pending is not None:
                            pending.continues = True
                            yield pending
                        raise
                    else:
                        if pending is not None:
                            yield pending

            else:

                @functools.wraps(function)
                async def wrapped(
                    *args: _P.args, **kwargs: _P.kwargs
                ) -> collections.abc.AsyncGenerator[AnnotatedResult, None]:
                    try:
                        async for result in asynciterfunction(*args, **kwargs):
                            if isinstance(result, AnnotatedResult):
                                yield result
                            else:
                                ares = make_result(result)
                                ares.continues = True
                                yield ares
                    except LastResult as exc:
                        if isinstance(exc.value, AnnotatedResult):
                            yield exc.value
                        else:
                            yield make_result(exc.value)

        elif more:
            iterfunction = typing.cast(
                collections.abc.Callable[_P, collections.abc.Iterable[_R]],
                function,
            )

            if delay_generator:

                @functools.wraps(function)
                def wrapped(
                    *args: _P.args, **kwargs: _P.kwargs
                ) -> collections.abc.Generator[AnnotatedResult, None, None]:
                    pending = None
                    try:
                        for result in iterfunction(*args, **kwargs):
                            if pending is not None:
                                pending.continues = True
                                yield pending
                                pending = None
                            if isinstance(result, AnnotatedResult):
                                yield result
                            else:
                                pending = make_result(result)
                    except LastResult as exc:
                        if pending is not None:
                            pending.continues = True
                            yield pending
                        if isinstance(exc.value, AnnotatedResult):
                            yield exc.value
                        else:
                            yield make_result(exc.value)
                    except:
                        if pending is not None:
                            pending.continues = True
                            yield pending
                        raise
                    else:
                        if pending is not None:
                            yield pending

            else:

                @functools.wraps(function)
                def wrapped(
                    *args: _P.args, **kwargs: _P.kwargs
                ) -> collections.abc.Generator[AnnotatedResult, None, None]:
                    try:
                        for result in iterfunction(*args, **kwargs):
                            if isinstance(result, AnnotatedResult):
                                yield result
                            else:
                                ares = make_result(result)
                                ares.continues = True
                                yield ares
                    except LastResult as exc:
                        if isinstance(exc.value, AnnotatedResult):
                            yield exc.value
                        else:
                            yield make_result(exc.value)

        elif asynchronous:
            asyncfunction = typing.cast(
                collections.abc.Callable[_P, collections.abc.Awaitable[_R]],
                function,
            )

            @functools.wraps(function)
            async def wrapped(
                *args: _P.args, **kwargs: _P.kwargs
            ) -> AnnotatedResult:
                try:
                    result = await asyncfunction(*args, **kwargs)
                except LastResult as exc:
                    result = exc.value
                if isinstance(result, AnnotatedResult):
                    return result
                return make_result(result)

        else:

            @functools.wraps(function)
            def wrapped(
                *args: _P.args, **kwargs: _P.kwargs
            ) -> AnnotatedResult:
                try:
                    result = function(*args, **kwargs)
                except LastResult as exc:
                    result = exc.value
                if isinstance(result, AnnotatedResult):
                    return result
                return make_result(result)

        # Employ setattr instead of directly setting it as that makes mypy and
        # pylint a lot happier.
        setattr(wrapped, "_varlink_signature", vlsig)
        return wrapped

    if function is None:
        # The decorator is called with parens.
        return wrap
    return wrap(function)


def varlinksignature(
    method: collections.abc.Callable[_P, _R],
) -> VarlinkMethodSignature | None:
    """Return the signature object constructed by the varlinkmethod decorator
    if the given method has been decorated.
    """
    # Wrap the access in getattr as that makes mypy and pylint a lot happier.
    return getattr(method, "_varlink_signature", None)


def _doc_to_comment(doc: str) -> typing.Iterator[str]:
    """Turn a doc string into a varlink IDL compatible comment."""
    return map("# ".__add__, inspect.cleandoc(doc).rstrip("\r\n").splitlines())


class VarlinkInterface:
    """A base class for varlink interface implementations.

    Deriving classes should set the name class attribute to the interface name
    and mark the interface methods with the varlinkmethod decorator.
    """

    name: str
    """The name of the varlink interface in dotted reverse domain notation."""

    errors: collections.abc.Sequence[type[TypedVarlinkErrorReply]] = ()
    """Optionally, an interface may list the possible errors that may be
    returned.

    For use with a server, this causes the errors with matching interface to be
    included in the rendered interface description. For use with a client
    proxy, this causes the error classes to be used for decoding error replies.
    """

    _error_map: dict[str, type[TypedVarlinkErrorReply]]
    """Constructed by __init_subclass__ from the errors sequence."""

    def __init_subclass__(
        cls: type["VarlinkInterface"], *, name: str | None = None
    ) -> None:
        try:
            cls.name
        except AttributeError:
            if name is None:
                raise RuntimeError(
                    "VarlinkInterface subclasses must define an interface name"
                ) from None
            cls.name = name
        else:
            if name is not None:
                raise RuntimeError(
                    "cannot specify VarlinkInterface name both via "
                    "inheritance and attribute"
                )
        try:
            validate_interface(cls.name)
        except ValueError as err:
            raise RuntimeError("invalid VarlinkInterface name") from err

        error_map = {}
        for error in cls.errors:
            if error.name in error_map:
                raise RuntimeError(f"duplicate error {error.name}")
            error_map[error.name] = error
        cls._error_map = error_map

    @classmethod
    def _generate_interface_description(
        cls, comments: bool
    ) -> typing.Iterator[str]:
        if comments and cls.__doc__:
            yield from _doc_to_comment(cls.__doc__)
        yield f"interface {cls.name}"

        typedefs: dict[str, str] = {}
        methods: list[str] = []
        for name in dir(cls):
            obj = getattr(cls, name)
            if (signature := varlinksignature(obj)) is None:
                continue
            _merge_typedefs(
                typedefs,
                signature.parameter_type.typedefs,
                signature.return_type.typedefs,
            )
            if comments and obj.__doc__:
                methods.append("")
                methods.extend(_doc_to_comment(obj.__doc__))
            methods.append(
                f"method {name}{signature.parameter_type.as_varlink} -> "
                f"{signature.return_type.as_varlink}"
            )

        errors: list[str] = []
        prefix = cls.name + "."
        for error in cls.errors:
            name = error.name.removeprefix(prefix)
            if name != error.name:
                _merge_typedefs(typedefs, error.paramtype.typedefs)
                if comments and error.__doc__:
                    errors.append("")
                    errors.extend(_doc_to_comment(error.__doc__))
                errors.append(f"error {name} {error.paramtype.as_varlink}")

        if typedefs or not comments:
            yield ""
        for tname, tdef in typedefs.items():
            yield f"type {tname} {tdef}"
        if typedefs and methods and not comments:
            yield ""
        yield from methods
        if methods and errors and not comments:
            yield ""
        yield from errors
        if methods or errors:
            yield ""

    @classmethod
    def render_interface_description(cls, *, comments: bool = True) -> str:
        """Render a varlink interface description from this interface.
        Refer to https://varlink.org/Interface-Definition. The comments
        parameter indicates whether __doc__ strings should be included as
        comments.
        """
        return "\n".join(cls._generate_interface_description(comments))

    @classmethod
    def raise_error(cls, reply: VarlinkMethodReply) -> typing.NoReturn:
        """Raise an exception from the given reply. It must be an error reply.
        If the error name is found in the errors sequence a
        TypedVarlinkErrorReply subclass is constructed. In that case, the given
        parameters are validated and may raise a ConversionError.
        """
        if reply.error is None:
            raise RuntimeError("raise_error invoked on non-error reply")
        try:
            errcls = cls._error_map[reply.error]
        except KeyError:
            raise GenericVarlinkErrorReply(
                reply.error, reply.parameters
            ) from None
        raise errcls(errcls.paramtype.fromjson(reply.parameters))



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

"""Basic type definitions."""

import asyncio
import collections.abc
import contextlib
import os
import re
import typing
import warnings


__all__ = [
    "FileDescriptor",
    "FileDescriptorArray",
    "FutureCounted",
    "HasFileno",
    "JSONObject",
    "JSONValue",
    "override",
    "validate_interface",
    "validate_name",
]

if typing.TYPE_CHECKING:
    from typing import override
else:

    def override(func):
        """typing.override stub to be removed when dropping 3.11 support."""
        return func


JSONValue = typing.Union[
    None, bool, float, int, str, list["JSONValue"], "JSONObject"
]


JSONObject = dict[str, JSONValue]


# pylint: disable=too-few-public-methods  # It's that one method we describe.
class HasFileno(typing.Protocol):
    """A typing protocol representing a file-like object and looking up the
    underlying file descriptor.

    If it has a close method, it must be compatible with the HasClose protocol.
    """

    def fileno(self) -> int:
        """Return the underlying file descriptor. When invoked on a closed
        object, the method may return a negative number or raise a
        ValueError.
        """


class HasClose(typing.Protocol):
    """A typing protocol representing objects that can be closed."""

    def close(self) -> None:
        """Can be called at most once to release resources."""


class HasFilenoAndClose(HasFileno, HasClose, typing.Protocol):
    """A file-like object supporting both fileno and close."""


def get_fileno(fdlike: int | HasFileno | None) -> int:
    """Return the underlying file descriptor number of an object that may
    represent an open file descriptor. Raises a ValueError in case there is no
    valid file descriptor (e.g. None).
    """
    if fdlike is None:
        raise ValueError("explicitly invalid file descriptor")
    try:
        fileno_meth = getattr(fdlike, "fileno")
    except AttributeError:
        if not isinstance(fdlike, int):
            raise TypeError(
                f"not a file descriptor: {type(fdlike).__name__}"
            ) from None
        if fdlike < 0:
            raise ValueError("negative file descriptor") from None
        return fdlike
    fdlike = fileno_meth()  # propagate ValueError
    if fdlike is None:
        # This is not covered by the protocol, but we can handle it.
        raise ValueError("fileno returned None")
    if not isinstance(fdlike, int):
        raise TypeError(f"fileno returned a non-int: {type(fdlike).__name__}")
    if fdlike < 0:
        raise ValueError("fileno() returned negative")
    return fdlike


class FileDescriptor:
    """Wrap various file descriptor objects of different types in a
    recognizable type for using in a varlink interface.
    """

    def __init__(self, fd: int | HasFileno | None, should_close: bool = False):
        """Wrap a file descriptor object that may be one of an integer, a
        higher level object providing a fileno method or None representing a
        invalid or closed file descriptor. Optionally, if the should_close flag
        is set and the FileDescriptor object is garbage collected without being
        closed, a warning is logged.
        """

        self.should_close: bool
        self.fd: int | HasFileno | None
        if isinstance(fd, int) and fd < 0:
            fd = None
        elif isinstance(fd, FileDescriptor):
            if should_close and fd.should_close:
                raise RuntimeError(
                    "FileDescriptor references another FileDescriptor and "
                    "both are flagged should_close"
                )
            self.fd = fd.fd
        else:
            self.fd = fd
        self.should_close = should_close

    @classmethod
    def make_pipe(
        cls, inheritable: bool = False, blocking: bool = False
    ) -> tuple["FileDescriptor", "FileDescriptor"]:
        """Create a pipe and wrap its ends in the FileDescriptor class."""
        flags = 0
        if not inheritable:
            flags |= os.O_CLOEXEC
        if not blocking:
            flags |= os.O_NONBLOCK
        if flags:
            piper, pipew = os.pipe2(flags)
        else:
            piper, pipew = os.pipe()
        return cls(piper), cls(pipew)

    def __bool__(self) -> bool:
        """Indicate whether the object refers to a plausibly open file
        descriptor.
        """
        try:
            get_fileno(self.fd)
        except ValueError:
            return False
        return True

    def fileno(self) -> int:
        """Return the underlying file descriptor, i.e. self. Raises a
        ValueError when closed.
        """
        return get_fileno(self.fd)

    __int__ = fileno

    def close(self) -> None:
        """Close the file descriptor. Idempotent. If the underlying file
        descriptor has a close method, it is used. Otherwise, os.close is used.
        """
        if self.fd is None:
            return
        try:
            try:
                close = getattr(self.fd, "close")
            except AttributeError:
                os.close(self.fileno())
            else:
                close()
        finally:
            self.fd = None

    def release(self) -> None:
        """Close the file descriptor if it should be closed. Do nothing
        otherwise.
        """
        if self.should_close:
            self.close()

    def __eq__(self, other: typing.Any) -> bool:
        """Compare two file descriptors. Comparison to integers, None or
        objects with a fileno method may succeed. Ownership is not considered
        for comparison.
        """
        try:
            selffileno = self.fileno()
        except ValueError:
            selffileno = None
        try:
            otherfileno = get_fileno(other)
        except TypeError:
            return False
        except ValueError:
            return selffileno is None
        return selffileno == otherfileno

    def __enter__(self) -> typing.Self:
        """Implement the context manager protocol yielding self and closing
        the file descriptor on exit. The object will be marked as being closed.
        """
        self.should_close = True
        return self

    def __exit__(self, *_: object) -> None:
        """Close the file descriptor on context manager exit."""
        self.close()

    def take(self) -> HasFileno | int | None:
        """Return and disassociate the wrapped file descriptor object. The
        FileDescriptor must be responsible for closing and thus marked with
        the should_close flag. This responsibility is transferred to the
        caller.
        """
        if not self.should_close:
            warnings.warn(
                f"unowned FileDescriptor {self.fd!r} being taken",
                ResourceWarning,
                stacklevel=2,
                source=self,
            )
        try:
            return self.fd
        finally:
            self.fd = None

    def __del__(self) -> None:
        """If the FileDescriptor is marked with the should_close flag, close it
        on garbage collection and issue a warning about closing it explicitly.
        """
        if self.fd is None or not self.should_close:
            return
        warnings.warn(
            f"owned FileDescriptor {self.fd!r} was never closed explicitly",
            ResourceWarning,
            source=self,
        )
        self.close()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.fd!r})"


class FutureCounted:
    """A reference counting base class. References are not simply counted.
    Instead referees are tracked individually. Any referee must be released
    eventually by calling release. Once all referees are gone, the destroy
    method is called once.
    """

    def __init__(self, initial: typing.Any) -> None:
        """The constructor consumes an initial referee. Otherwise, it would be
        immediately destroyed.
        """
        self._references: set[int] = {id(initial)}

    @classmethod
    @contextlib.contextmanager
    def new_managed(cls) -> collections.abc.Iterator[typing.Self]:
        """Create a managed object whose lifetime is managed in a context
        manager.
        """
        sentinel = object()
        obj = cls(sentinel)
        try:
            yield obj
        finally:
            obj.release(sentinel)

    def reference(self, referee: typing.Any) -> None:
        """Record an object as referee. The referee should be either passed to
        release once or garbage collected by Python.
        """
        if not self._references:
            raise RuntimeError("cannot reference destroyed object")
        objid = id(referee)
        assert objid not in self._references
        self._references.add(objid)

    def reference_until_done(self, fut: asyncio.Future[typing.Any]) -> None:
        """Reference this object until the passed future is done."""
        self.reference(fut)
        fut.add_done_callback(self.release)

    def release(self, referee: typing.Any) -> None:
        """Release the reference identified by the given referee. If this was
        the last reference, this object is destroyed. Releasing a referee that
        was not referenced is an error as is releasing a referee twice.
        """
        objid = id(referee)
        try:
            self._references.remove(objid)
        except KeyError:
            raise RuntimeError(
                f"releasing reference to unregistered object {referee!r}"
            ) from None
        if not self._references:
            self.destroy()

    def destroy(self) -> None:
        """Called when the last reference is released."""
        raise NotImplementedError


class FileDescriptorArray(FutureCounted):
    """Represent an array of observed or owned file descriptors. When the array
    is released, the contained file descriptors are released which amounts to
    closing the owned ones. The lifetime can be controlled in two ways.
    Responsibility for individual file descriptors can be assumed by using the
    take method and thus removing them from the array. The lifetime of the
    entire array can be extended using the FutureCounted mechanism inherited
    from. Each file descriptor in the array must be unique.
    """

    def __init__(
        self,
        initial_referee: typing.Any,
        fds: collections.abc.Iterable[HasFileno | int | None] | None = None,
    ):
        """Construct a FileDescriptorArray. The initial_referee is passed to
        the FutureCounted constructor. The fds iterable must not contain any
        fileno several times. Note that file descriptors not wrapped in
        FileDescriptor are assumed to be owned and shall be closed.
        """
        super().__init__(initial_referee)
        self._by_position: list[FileDescriptor] = []
        self._by_fileno: dict[int, int] = {}
        for fdlike in fds or ():
            try:
                fileno = get_fileno(fdlike)
            except ValueError:
                fileno = None
            if fileno is None:
                self._by_position.append(FileDescriptor(None))
            elif fileno in self._by_fileno:
                raise ValueError(
                    f"file descriptor {fileno} passed several times"
                )
            else:
                fd = FileDescriptor(fdlike, should_close=True)
                if fileno is not None:
                    self._by_fileno[fileno] = len(self._by_position)
                self._by_position.append(fd)

    def __bool__(self) -> bool:
        """Are there any owned file descriptors in the array?"""
        return any(self._by_position)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FileDescriptorArray):
            return False
        if len(self._by_position) != len(other._by_position):
            return False
        return all(
            fd1 == fd2
            for fd1, fd2 in zip(self._by_position, other._by_position)
        )

    def __len__(self) -> int:
        return len(self._by_position)

    def __getitem__(self, index: int) -> FileDescriptor:
        return self._by_position[index]

    def __iter__(self) -> collections.abc.Iterator[FileDescriptor]:
        return iter(self._by_position)

    @override
    def destroy(self) -> None:
        """Release all file descriptors. This amounts to closing owned ones.
        Idempotent.
        """
        for fd in self:
            fd.release()

    def add(self, fdlike: HasFileno | int | None) -> int:
        """Append a file descriptor to the array and return its position. If
        None or a closed file descriptor are passed, a ValueError is raised.
        If the file descriptor already is in the array, the add call must pass
        the same representation (equal integer or identical object) and the
        existing index is returned. Descriptors not wrapped in the
        FileDescriptor class are closed eventually.
        """
        fileno = get_fileno(fdlike)  # Propagate ValueError and TypeError
        try:
            index = self._by_fileno[fileno]
        except KeyError:
            if not isinstance(fdlike, FileDescriptor):
                fdlike = FileDescriptor(fdlike, True)
            index = len(self._by_position)
            self._by_position.append(fdlike)
            self._by_fileno[fileno] = index
            return index
        other = self._by_position[index]
        if other is fdlike or other.fd is fdlike or isinstance(fdlike, int):
            return index
        raise ValueError(
            f"attempt to add {fileno} with different representation "
            f"{fdlike!r} than existing entry {other.fd!r}"
        )

    def __del__(self) -> None:
        fds = [fd for fd in self._by_position if fd.should_close and fd]
        if not fds:
            return
        warnings.warn(
            "FileDescriptorArray not destroyed before garbage collection",
            ResourceWarning,
            source=self,
        )
        for fd in fds:
            fd.release()


def validate_interface(interface: str) -> None:
    """Validate a varlink interface in reverse-domain notation. May raise a
    ValueError.
    """
    if not re.match(
        r"[A-Za-z](?:-*[A-Za-z0-9])*(?:\.[A-Za-z0-9](?:-*[A-Za-z0-9])*)+",
        interface,
    ):
        raise ValueError(f"invalid varlink interface {interface!r}")


def validate_name(name: str) -> None:
    """Validate a varlink name. May raise a ValueError."""
    if not re.match(r"^[A-Z][A-Za-z0-9]*$", name):
        raise ValueError(f"invalid varlink name {name!r}")



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

"""Represent a varlink error values in Python."""

import functools
import typing

from .conversion import ObjectVarlinkType, VarlinkType
from .message import VarlinkMethodReply
from .types import JSONObject, JSONValue, override


class _VarlinkErrorParameterDescriptor:
    def __init__(self, name: str):
        # Usually, this would be __set_name__, but as we assign with setattr,
        # we handle the name propagation in the ctor.
        self.name = name

    def __get__(self, obj: "TypedVarlinkErrorReply", owner: type) -> JSONValue:
        try:
            # mypy does not know that obj.parameters is a JSONObject.
            return typing.cast(JSONValue, obj.parameters[self.name])
        except KeyError as err:
            raise AttributeError(f"no attribute named {self.name}") from err

    def __set__(self, obj: "TypedVarlinkErrorReply", value: JSONValue) -> None:
        obj.parameters[self.name] = value


class VarlinkErrorReply(Exception):
    """Base class for protocol-level errors communicated via Varlink."""

    name: str
    """The name of the error prefixed with the interface name in reverse-domain
    notation. This may be a class attribute or an instance attribute.
    """

    def paramstojson(self) -> JSONObject:
        """Represent the parameters as a JSONObject."""
        raise NotImplementedError

    def toreply(self) -> VarlinkMethodReply:
        """Represent the error as a VarlinkMethodReply."""
        return VarlinkMethodReply(self.paramstojson(), error=self.name)

    def tojson(self) -> JSONObject:
        """Represent the entire error as a JSONObject."""
        return self.toreply().tojson()


class GenericVarlinkErrorReply(VarlinkErrorReply):
    """An untyped VarlinkErrorReply subclass representing errors for which no
    TypedVarlinkErrorReply subclass exists.
    """

    def __init__(self, error: str, parameters: JSONObject | None = None):
        """In general, a varlink error object consists of a fully qualified
        error name and a parameter object and that's what should be supplied
        here.
        """
        super().__init__()
        self.name = error
        self.parameters = {} if parameters is None else parameters

    @override
    def paramstojson(self) -> JSONObject:
        return self.parameters

    def __repr__(self) -> str:
        if not self.parameters:
            return f"{self.__class__.__name__}({self.name!r})"
        return f"{self.__class__.__name__}({self.name!r}, {self.parameters!r})"


@typing.dataclass_transform()
class TypedVarlinkErrorReply(VarlinkErrorReply):
    """Class hierarchy for type-driven protocol-level errors communicated via
    Varlink. The name attribute on this hierarchy is a class attribute and not
    an instance attribute. It may be specified in one of the following ways.
     * Set the name class attribute to a fully qualified name.
     * Pass a fully qualified name as keyword argument name for class
       inheritance.
     * Pass a qualified interface as keyword argument interface for class
       inheritance. In this case the actual name will be constructed from
       the interface and the name of the derived class.
    Each subclass must provide a parameter specification via one of the
    following means.
     * Annotate the attribute parameters as a typing.TypedDict.
     * Create an inner class named Parameters whose attribute annotations
       are used to construct a typing.TypedDict.
    A suitable __init__ method setting the parameters instance attribute will
    be created automatically. For each parameter, a matching descriptor
    attribute will be added. The prefix of these attributes may be customized
    via the paramprefix keyword argument for class inheritance and defaults to
    "p_". If no parameters clash with keywords or methods, it is suggested to
    request an empty paramprefix.
    """

    paramtype: ObjectVarlinkType
    """A VarlinkType instance representing the type of the parameters. It can
    be used for converting a Python representation and a JSON representation
    of the parameters back and forth. It is constructed by the
    __init_subclass__ hook from the annotations of the parameters attribute or
    the Parameters class attribute.
    """

    def __init_subclass__(
        cls: type["TypedVarlinkErrorReply"], **kwargs: typing.Any
    ) -> None:
        set_name = False
        try:
            name = cls.name
        except AttributeError:
            if "name" in kwargs:
                name = kwargs.pop("name")
                set_name = True
            elif "interface" in kwargs:
                name = kwargs.pop("interface") + "." + cls.__name__
                set_name = True
            else:
                raise RuntimeError(
                    "For instantiating a TypedVarlinkErrorReply subclass you "
                    "must set on of the name class attribute, name class "
                    "keyword argument or interface class keyword argument."
                ) from None
        paramprefix = kwargs.pop("paramprefix", "p_")
        super().__init_subclass__(**kwargs)
        if set_name:
            cls.name = name
        try:
            paramtype = cls.__annotations__["parameters"]
        except KeyError:
            try:
                paramcls = getattr(cls, "Parameters")
            except AttributeError:
                raise RuntimeError(
                    "For instantiating a TypedVarlinkErrorReply subclass you "
                    "must provide a type annotation for the parameters "
                    "attribute or add a type annotated inner class named "
                    "Parameters."
                ) from None
            # mypy does not like to construct a TypedDict with non-constant
            # arguments, but that's exactly what we're up to here. It also
            # objects to assigning a type hint to a variable assigned elsewhere
            # as a TypedDict may already be defined as an annotation of the
            # the parameters attribute. Silence both.
            paramtype = typing.TypedDict(  # type: ignore[misc,no-redef]
                f"{cls.__name__}Parameters", paramcls.__annotations__
            )
            cls.__annotations__["parameters"] = paramtype
        else:
            if not typing.is_typeddict(paramtype):
                raise RuntimeError(
                    "The parameters attribute must be annotated as a "
                    "typing.TypedDict."
                )
        vtype = VarlinkType.from_type_annotation(paramtype)
        # As we pass a TypedDict, we actually get an ObjectVarlinkType.
        assert isinstance(vtype, ObjectVarlinkType)
        cls.paramtype = vtype

        # The type annotation given here does not really matter as we override
        # it in the next step.
        @functools.wraps(cls.__init__)
        def subclass_init(
            self: typing.Any,
            _mapping: JSONObject | None = None,
            /,
            **kwargs: JSONValue,
        ) -> None:
            super().__init__()
            if _mapping is None:
                self.parameters = kwargs
            else:
                self.parameters = _mapping | kwargs

        subclass_init.__annotations__ = {
            "_mapping": paramtype,
        } | paramtype.__annotations__
        setattr(cls, "__init__", subclass_init)
        for key, ann in paramtype.__annotations__.items():
            desc = _VarlinkErrorParameterDescriptor(key)
            descname = paramprefix + key
            setattr(cls, descname, desc)
            cls.__annotations__[descname] = ann

    @override
    def paramstojson(self) -> JSONObject:
        ret = self.paramtype.tojson(self.parameters)
        assert isinstance(ret, dict)  # self.parameters is a dict.
        return ret

    if typing.TYPE_CHECKING:
        # Pretend that we have an __init__ as mypy does not recognize the
        # generated __init__ method.
        def __init__(
            self, _mapping: JSONObject | None = None, /, **_kwargs: JSONValue
        ): ...

        # Pretend that we have __getattr__ such that mypy and pylint do not
        # moan about accessing our descriptor objects. Unfortunately, this
        # entirely breaks checking attributes.
        def __getattr__(self, attr: str) -> typing.Any: ...



# Copyright 2026 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

import asyncio
import os
import socket
import unittest.mock


def async_read_fd(fd: int, size: int) -> asyncio.Future[bytes]:
    loop = asyncio.get_running_loop()
    fut = loop.create_future()

    def readable() -> None:
        loop.remove_reader(fd)
        try:
            data = os.read(fd, size)
        except Exception as exc:
            fut.set_exception(exc)
        else:
            fut.set_result(data)

    loop.add_reader(fd, readable)
    return fut


def _send_fds_writable(
    sock: socket.socket, data: bytes, fds: list[int], fut: asyncio.Future[None]
) -> None:
    loop = asyncio.get_running_loop()
    try:
        sent = socket.send_fds(sock, [data], fds)
    except Exception as exc:
        loop.remove_writer(sock)
        fut.set_exception(exc)
    else:
        if sent >= len(data):
            loop.remove_writer(sock)
            fut.set_result(None)
        else:
            loop.add_writer(
                sock, _send_fds_writable, sock, data[sent:], [], fut
            )


def async_send_fds(
    sock: socket.socket, data: bytes, fds: list[int] | None = None
) -> asyncio.Future[None]:
    loop = asyncio.get_running_loop()
    if fds is None:
        fds = []
    fut = loop.create_future()
    loop.add_writer(sock, _send_fds_writable, sock, data, fds, fut)
    return fut


async def defer(
    count: int = 100, until_called: unittest.mock.Mock | None = None
) -> None:
    for _ in range(count):
        if until_called is not None and until_called.called:
            return
        await asyncio.sleep(0)



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

import asyncio
import contextlib
import os
import socket
import unittest
from unittest.mock import Mock

from asyncvarlink import (
    JSONObject,
    JSONValue,
    VarlinkBaseProtocol,
    VarlinkMethodCall,
    VarlinkMethodReply,
    VarlinkProtocol,
    VarlinkTransport,
)

from helpers import defer


class TransportTests(unittest.IsolatedAsyncioTestCase):
    async def test_receive_socket(self) -> None:
        protocol = VarlinkBaseProtocol()
        protocol.message_received = Mock(return_value=None)
        protocol.connection_lost = Mock(return_value=None)
        sock1, sock2 = socket.socketpair()
        with contextlib.closing(sock1):
            with contextlib.closing(sock2):
                transport = VarlinkTransport(
                    asyncio.get_running_loop(), sock1, sock1, protocol
                )
                protocol.eof_received = transport.close
                sock2.send(b"hello")
                await defer(until_called=protocol.message_received)
                protocol.message_received.assert_called_once_with(
                    b"hello", None
                )
            await defer(until_called=protocol.connection_lost)
            protocol.connection_lost.assert_called_once_with(None)

    async def test_receive_socket_eof(self) -> None:
        protocol = VarlinkBaseProtocol()
        protocol.eof_received = Mock()
        sock1, sock2 = socket.socketpair()
        with contextlib.closing(sock1), contextlib.closing(sock2):
            VarlinkTransport(
                asyncio.get_running_loop(), sock1, sock1, protocol
            )
            sock2.close()
            await defer(until_called=protocol.eof_received)
        protocol.eof_received.assert_called_once_with()

    async def test_receive_pipe(self) -> None:
        protocol = VarlinkBaseProtocol()
        protocol.message_received = Mock(return_value=None)
        pipe1, pipe2 = os.pipe()
        try:
            VarlinkTransport(
                asyncio.get_running_loop(), pipe1, pipe2, protocol
            )
            os.write(pipe2, b"hello")
            await defer(until_called=protocol.message_received)
        finally:
            os.close(pipe2)
            os.close(pipe1)
        protocol.message_received.assert_called_once_with(b"hello", None)

    async def test_receive_pipe_eof(self) -> None:
        protocol = VarlinkBaseProtocol()
        protocol.eof_received = Mock()
        pipe1, pipe2 = os.pipe()
        VarlinkTransport(asyncio.get_running_loop(), pipe1, pipe2, protocol)
        os.close(pipe2)
        await defer(until_called=protocol.eof_received)
        protocol.eof_received.assert_called_once_with()

    async def test_send_socket(self) -> None:
        loop = asyncio.get_running_loop()
        protocol = VarlinkBaseProtocol()
        sock1, sock2 = socket.socketpair(
            type=socket.SOCK_STREAM | socket.SOCK_NONBLOCK
        )
        with contextlib.closing(sock1), contextlib.closing(sock2):
            transport = VarlinkTransport(loop, sock1, sock1, protocol)
            fut1 = transport.send_message(b"hello")
            fut2 = transport.send_message(b"world")
            self.assertEqual(b"helloworld", await loop.sock_recv(sock2, 1024))
            self.assertIsNone(fut1.result())
            self.assertIsNone(fut2.result())
            transport.close()
            fut3 = transport.send_message(b"fail")
            with self.assertRaises(OSError):
                fut3.result()


class ProtocolTests(unittest.IsolatedAsyncioTestCase):
    async def test_receive(self) -> None:
        protocol = VarlinkProtocol()
        pipe1, pipe2 = os.pipe()
        try:
            VarlinkTransport(
                asyncio.get_running_loop(), pipe1, pipe2, protocol
            )
            protocol.request_received = Mock()
            protocol.message_received(b'{"hello":"world"}\0', None)
            await defer(until_called=protocol.request_received)
            protocol.request_received.assert_called_once_with(
                {"hello": "world"}, None
            )
        finally:
            os.close(pipe2)
            os.close(pipe1)

    async def test_receive_error(self) -> None:
        protocol = VarlinkProtocol()
        pipe1, pipe2 = os.pipe()
        try:
            VarlinkTransport(
                asyncio.get_running_loop(), pipe1, pipe2, protocol
            )
            protocol.error_received = Mock(wraps=protocol.error_received)
            protocol.message_received(b"}\0", None)
            await defer(until_called=protocol.error_received)
            protocol.error_received.assert_called_once()
        finally:
            os.close(pipe2)
            os.close(pipe1)

    async def test_receive_pause(self) -> None:
        loop = asyncio.get_running_loop()
        protocol = VarlinkProtocol()
        pipe1, pipe2 = os.pipe()
        try:
            transport = VarlinkTransport(loop, pipe1, pipe2, protocol)
            futs = [loop.create_future(), loop.create_future()]
            protocol.request_received = Mock(side_effect=futs)
            await asyncio.sleep(0)
            self.assertFalse(transport._paused)
            protocol.message_received(b'{"a":0}\0{"b":0}\0', None)
            await asyncio.sleep(0)
            self.assertTrue(transport._paused)
            protocol.request_received.assert_called_once_with({"a": 0}, None)
            self.assertTrue(transport._paused)
            futs[0].set_result(None)
            await asyncio.sleep(0)
            self.assertTrue(transport._paused)
            protocol.request_received.assert_called_with({"b": 0}, None)
            await asyncio.sleep(0)
            self.assertTrue(transport._paused)
            futs[1].set_result(None)
            await asyncio.sleep(0)
            self.assertFalse(transport._paused)
        finally:
            os.close(pipe2)
            os.close(pipe1)

    async def test_receive_multiple(self) -> None:
        protocol = VarlinkProtocol()
        pipe1, pipe2 = os.pipe()
        try:
            VarlinkTransport(
                asyncio.get_running_loop(), pipe1, pipe2, protocol
            )
            protocol.request_received = Mock()
            await asyncio.sleep(0)
            protocol.message_received(b'{"a":0}\0{"b":0}\0', None)
            await defer(until_called=protocol.request_received)
            protocol.request_received.assert_called_once_with({"a": 0}, None)
            await asyncio.sleep(0)
            protocol.request_received.assert_called_with({"b": 0}, None)
        finally:
            os.close(pipe2)
            os.close(pipe1)

    async def test_send(self) -> None:
        loop = asyncio.get_running_loop()
        protocol = VarlinkProtocol()
        sock1, sock2 = socket.socketpair(
            type=socket.SOCK_STREAM | socket.SOCK_NONBLOCK
        )
        with contextlib.closing(sock1), contextlib.closing(sock2):
            VarlinkTransport(loop, sock1, sock1, protocol)
            await asyncio.sleep(0)
            protocol.send_message({"hello": "world"}, [])
            self.assertEqual(
                b'{"hello":"world"}\0', await loop.sock_recv(sock2, 1024)
            )


class CallReplyTest(unittest.TestCase):
    def test_valid_call(self) -> None:
        cases: list[JSONObject] = [
            {},
            {"parameters": {"ingredient": "egg"}},
            {"oneway": True},
            {"more": True},
            {"upgrade": True},
            {"more": True},
            {"com.example.Extension": "spam"},
        ]
        for jobj in cases:
            jobj["method"] = "com.example.Spam"
            call = VarlinkMethodCall.fromjson(jobj)
            self.assertEqual(jobj, call.tojson())

    def test_invalid_call(self) -> None:
        cases: list[tuple[type[Exception], JSONValue]] = [
            (TypeError, 42),
            (ValueError, {}),
            (TypeError, {"method": 42}),
            (ValueError, {"method": "Unqualified"}),
            (ValueError, {"method": "com.example.lowercase"}),
            (TypeError, {"method": "com.example.Spam", "oneway": "egg"}),
            (TypeError, {"method": "com.example.Spam", "more": "egg"}),
            (TypeError, {"method": "com.example.Spam", "upgrade": "egg"}),
            (
                ValueError,
                {"method": "com.example.Spam", "more": True, "oneway": True},
            ),
            (TypeError, {"method": "com.example.Spam", "parameters": "egg"}),
        ]
        for exc, jobj in cases:
            with self.assertRaises(exc):
                VarlinkMethodCall.fromjson(jobj)

    def test_valid_reply(self) -> None:
        cases: list[JSONObject] = [
            {},
            {"error": "com.example.Spam"},
            {"parameters": {"ingredients": "egg"}},
            {"continues": True},
        ]
        for jobj in cases:
            reply = VarlinkMethodReply.fromjson(jobj)
            self.assertEqual(jobj, reply.tojson())

    def test_invalid_reply(self) -> None:
        cases: list[tuple[type[Exception], JSONValue]] = [
            (TypeError, 42),
            (TypeError, {"error": 42}),
            (ValueError, {"error": "Unqualified"}),
            (ValueError, {"error": "com.example.lowercase"}),
            (TypeError, {"parameters": "egg"}),
            (TypeError, {"continues": "egg"}),
        ]
        for exc, jobj in cases:
            with self.assertRaises(exc):
                VarlinkMethodReply.fromjson(jobj)
